
//  ------------------------------------------------------------------
//  Goldware Binary Difference and Patching Utilities.
//  Copyright (C) Odinn Sorensen. Released as FREEWARE.
//  Based on the BASIC sourcecode of JLPAK/UNPAK 1.0 by John Larmouth.
//  ------------------------------------------------------------------
//  Binary patching utility.
//  ------------------------------------------------------------------


#include <new.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>
#include <sys/types.h>
#include <sys/stat.h>
#if defined(__WATCOMC__)
#include <sys/utime.h>
#else
#include <utime.h>
#endif


//  ------------------------------------------------------------------
//  Portability stuff

#define not   !
#define and   &&
#define or    ||
#define true  1
#define false 0

#define NUL   '\0'

typedef int boolean;
typedef unsigned long ulong;
typedef unsigned char byte;

#ifdef __GNUC__
#define ios__binary ios::bin
#else
#define ios__binary ios::binary
#endif


//  ------------------------------------------------------------------

// IBUF is the input buffer size for obtaining the ascii value of each
// byte. Speed seems relatively independent of the IBUF size.
const int IBUF = 100;

// This is the hunk of the old file that we work with. We read in the first
// 32000 (or all if less), and discard as commanded in the compressed format.
// This is a parameter of the FCM format.
const int ISIZE = 32000;

// This is the amount of FCM we do before moving it.
const int ISIZE2 = 24000;

// And this is the amount we move it down each time.
const int ISIZE3 = 12000;

// This is the size of type A and no match blocks, and is a parameter of the
// compressed (FCM) format.
const int ISEG = 16;

// This is the array of bytes from the old version
byte* bytesold;

// This is the array of bytes from the FCM format file
byte* bytesfcm;

// Filenames
char* trace = NULL;
char* oldfile = NULL;
char* gdffile = NULL;
char* newfile = NULL;
char* wrkfile = "gpatch.tmp";


//  ------------------------------------------------------------------

int bitstringptr;
int fcmpointer;

boolean getbit() {

  static int bitstring;

  if(bitstringptr == 0)
    bitstring = bytesfcm[fcmpointer++];
  else
    bitstring <<= 1;
  bitstringptr++;
  bitstringptr %= 8;
  return (bitstring & 0x80) ? true : false;
}


//  ------------------------------------------------------------------

ofstream fp;
ofstream dbg;
char* obuff;
char* obuffptr;
int obufflen = 0;
const int OBUF = 1000;

int sumcheck1 = 0;
int sumcheck2 = 0;
long outputcnt;

void outputsub(int octetvalue) {

  if(obufflen > OBUF) {
    fp.write(obuff, obufflen);
    obufflen = 0;
    obuffptr = obuff;
  }

  if(octetvalue < 0 or octetvalue > 255) {
    cout << "**** ERROR - Data format error - Octetvalue out of range (bug): " << octetvalue << endl;
    exit(-1);
  }

  sumcheck1 = (sumcheck1 + octetvalue) % 255;
  sumcheck2 = (sumcheck2 + sumcheck1) % 255;
  *obuffptr++ = (byte)octetvalue;
  obufflen++;
  outputcnt++;
}


//  ------------------------------------------------------------------

long donefcm;
long lengthfcm;
boolean quiet = false;

void progress() {

  if(not quiet) {
    long realperc = ((fcmpointer - 1 + donefcm) * 100L) / lengthfcm;
    cout << setw(3) << realperc << "%\b\b\b\b";
  }
}


//  ------------------------------------------------------------------

boolean help = false;
boolean gdffmt = true;

int main(int argc, char** argv) {

  char obuffer[OBUF+100];
  obuff = obuffer;

  #if defined(__BORLANDC__)
  set_new_handler(0);
  #endif

  int argn = 1;
  int gotargn = 1;
  while(argn < argc) {
    char* argp = argv[argn];
    if(*argp == '-') {
      switch(argp[1]) {
        case '?':
        case 'h':
          help = true;
          break;
        case 'o':
          gdffmt = false;
          break;
        case 'q':
          quiet = true;
          break;
        case 't':
          argn++;
          if(argn < argc)
            trace = argv[argn];
          break;
      }
    }
    else {
      switch(gotargn) {
        case 1:   // <gdffile>
          gdffile = argp;
          break;
        case 2:   // [oldfile]
          oldfile = argp;
          break;
        case 3:   // [newfile]
          newfile = argp;
          break;
        case 4:   // [wrkfile]
          wrkfile = argp;
          break;
      }
      gotargn++;
    }
    argn++;
  }

  if(not quiet or help) {
    cout << "GPATCH 1.0 - Binary patching utility. FREEWARE. Copyright (C) Odinn Sorensen." << endl;
    cout << "-----------------------------------------------------------------------------" << endl;
    cout << endl;
  }

  if(gotargn < 2 or help) {
    cout << "Syntax: gpatch [-options] <gdffile> [oldfile] [newfile] [wrkfile]" << endl;
    cout << endl;
    cout << "Options:" << endl;
    cout << "  -t <tracefile>  Writes debugging info to a trace file." << endl;
    cout << "  -q              Quiet - no screen output." << endl;
    cout << "  -o              Patch file is original JLPAK .FCM format." << endl;
    cout << "  -h or -?        This help text." << endl;
    cout << endl;
    cout << "This program is based on JLUNPAK by Prof J.Larmouth, U of Salford, England." << endl;
    cout << endl;
    cout << "Converted from the original BASIC sourcecode by Odinn Sorensen." << endl;
    cout << "Contact: odinn@goldware.dk, http://www.goldware.dk, FidoNet 2:236/77." << endl;
    exit(1);
  }

  // Allocate memory
  bytesold = new byte[ISIZE+1];
  bytesfcm = new byte[ISIZE+1];

  // This is the amount of the fcm file that we have already processed as a
  // previous hunk, and of the old file that we have used.
  long usedold = 0;
  donefcm = 0;

  ofstream dbg;
  if(trace)
    dbg.open(trace);

  ifstream fpfcm;

  fpfcm.open(gdffile, ios::in|ios__binary);
  if(not fpfcm) {
    cout << "**** ERROR - File " << gdffile << " not found!" << endl;
    exit(-1);
  }

  fpfcm.seekg(0, ios::end);
  lengthfcm = fpfcm.tellg();
  fpfcm.seekg(0);

  char buff[IBUF];
  memset(buff, ' ', IBUF);

  int ICNT2 = 0;
  while(not fpfcm.eof() and ICNT2 + IBUF <= ISIZE) {
    fpfcm.read(buff, IBUF);
    memcpy(bytesfcm+ICNT2+1, buff, IBUF);
    ICNT2 += IBUF;
  }

  int realendfcm = ISIZE;
  int endfcm = ISIZE2;
  if(lengthfcm < ISIZE) {
    endfcm = (int)lengthfcm;
    realendfcm = (int)lengthfcm;
  }

  fcmpointer = 1;

  char newfiledef[256];
  char oldfiledef[256];

  //  1. a magic code
  //  2. a length count (one octet) for the next item.
  //  3. the full path name of the file that is to be compressed
  //     (on decompression, a file TEMP.TMP will be created in the current
  //     directory, and then copied to the original destination path when
  //     decompression is complete, unless overridden by a different target).
  //  4. two octets of sum check for the whole of this file;  the sum check
  //     is the standard one of SIGMA(Ai) and SIGMA(iAi), all MOD 255.
  //  5. a length count for the next item.
  //  6. the full path name of the reference file (this can be overridden
  //     on decompression)
  //  7. two octets of sum check for the first 1K of the reference file.
  //  8. four octets of timestamp for the target file.

  if(gdffmt) {
    char magic1 = bytesfcm[fcmpointer++];
    char magic2 = bytesfcm[fcmpointer++];

    if(magic1 != 'l' or magic2 != 's') {
      cout << "**** ERROR - Unknown file format in " << gdffile << endl;
      exit(-1);
    }
  }

  int octlen = bytesfcm[fcmpointer];
  memcpy(newfiledef, bytesfcm+fcmpointer+1, octlen);
  newfiledef[octlen] = NUL;
  if(newfile == NULL)
    newfile = newfiledef;
  fcmpointer += octlen + 1;
  int sumcheckval1 = bytesfcm[fcmpointer++];
  int sumcheckval2 = bytesfcm[fcmpointer++];
  octlen = bytesfcm[fcmpointer];
  memcpy(oldfiledef, bytesfcm+fcmpointer+1, octlen);
  oldfiledef[octlen] = NUL;
  if(oldfile == NULL)
    oldfile = oldfiledef;
  fcmpointer += octlen + 1;
  int oldsumcheckval1 = bytesfcm[fcmpointer++];
  int oldsumcheckval2 = bytesfcm[fcmpointer++];
  ulong newstamp;
  if(gdffmt) {
    newstamp  =  (ulong)bytesfcm[fcmpointer++];
    newstamp |= ((ulong)bytesfcm[fcmpointer++]) << 8;
    newstamp |= ((ulong)bytesfcm[fcmpointer++]) << 16;
    newstamp |= ((ulong)bytesfcm[fcmpointer++]) << 24;
  }

  if(not quiet)
    cout << "Creating " << newfile << " from " << oldfile << " and " << gdffile << " ... ";

  ifstream fpold;
  fpold.open(oldfile, ios::in|ios__binary);
  if(not fpold) {
    //fpold.setbuf(NULL, 32000);
    cout << "**** ERROR - Reference file " << oldfile << " not found!" << endl;
    exit(-1);
  }

  long lengthold;
  int checked = 0;
  int oldsumcheck1 = 0;
  int oldsumcheck2 = 0;
  memset(buff, ' ', IBUF);
  while(not fpold.eof()) {
    fpold.read(buff, IBUF);
    if(checked < 10) {
      checked++;
      if(not fpold.eof()) {
        for(int i=0; i<IBUF; i++) {
          oldsumcheck1 = (oldsumcheck1 + buff[i]) % 255;
          oldsumcheck2 = (oldsumcheck2 + oldsumcheck1) % 255;
        }
      }
      else {
        lengthold = fpold.tellg();
        for(int i=0; i<(lengthold%IBUF); i++) {
          oldsumcheck1 = (oldsumcheck1 + buff[i]) % 255;
          oldsumcheck2 = (oldsumcheck2 + oldsumcheck1) % 255;
        }
      }
    }
    else
      break;
  }
  fpold.seekg(0, ios::end);
  lengthold = fpold.tellg();

  if(oldsumcheckval1 != oldsumcheck1 or oldsumcheckval2 != oldsumcheck2) {
    cout << "**** ERROR - Sumcheck of reference file differs from original" << endl;
    exit(-1);
  }

  fpold.seekg(0);

  int ICNT1 = 0;
  memset(buff, ' ', IBUF);
  while(not fpold.eof() and ICNT1 + IBUF <= ISIZE) {
    fpold.read(buff, IBUF);
    memcpy(bytesold+ICNT1+1, buff, IBUF);
    ICNT1 += IBUF;
  }

  int endold = ISIZE;
  if(lengthold < ISIZE)
    endold = (int)lengthold;

  // Now we initialise all variables, ready to start the first hunk
  // decompression

  // The receivers state is set up with ptroffset = 0, diffa = 0, and
  // diffb = 0 and donefcm& = 0 and usedold& = 0.
  // These values are retained unless explicitly
  // changed. The difference is new - old

  int i, ijl;

  // Used to hold the global difference values we maintain.
  int gdiff[4+1];

  for(i=1; i<=4; i++)
    gdiff[i] = 0;
  int ptroffset = 0;
  int cnt = 0;
  outputcnt = 0;
  obuffptr = obuff;
  obufflen = 0;
  boolean reallyfinished = false;
  boolean finished = false;
  int negate, octets, move, octval, copybytes, numocts;
  int diffindx, diffindx1, diffindx2, diffindx3, diffindx4;

  remove(wrkfile);
  fp.open(wrkfile, ios::out|ios__binary);
  //fp.setbuf(NULL, 32000);

  do {  // This is the outer loop, loop until finished, for each hunk.

    // We take each octet of fcmbytes in turn.

    while(fcmpointer <= endfcm) {

      progress();

      int code = bytesfcm[fcmpointer];

      // On decompression, we branch on bit 1 equals 1 or 0, and if bit 1 is
      // 1, then we branch on 10 or 11.

      int count;
      int codecase = 0;
      if(code >= 128) {
        if((code / 64) == 2) {
          //
          // codecase = 1 - no match
          // For 10, we pick up nnnnnn, and copy
          // ISEG * nnnnnn octetcs from the fcm to the target.
          //
          codecase = 1;   // Bits are 10 - no match for count segments
          count = code % 64;
          if(count == 0)
            count = 64;
        }
        else {
          //
          // codecase = 2 - total match (type A)
          // This is the branch from first bit 1, first two 11.
          // We copy ISEG * nnnnnn octets from the old file, starting at cnt +
          // ptroffset + 1. nnnnnn is in count.
          //
          codecase = 2;   // Bits are 11 - total match for count segments
          count = code % 64;
          if(count == 0)
            count = 64;
        }
      }
      else {
        //
        // codecase = 3 - pointer change
        // If there is a pointer change, we need to signal it in the output.
        // We have four cases for the value of the ptr change:
        //              positive and less than (or equals) 256
        //              negative and less than (or equals) 256
        //              positive and over 256
        //              negative and over 256
        // We use 00000000, 00000001, 00010000, and 00010001 for these four cases.
        // In the first two cases, we follow with a single octet, and in the last
        // two with two octets. In the case of the single octet, zero means 256.
        // On decompression, we are in the first bit zero case, and branch first on
        // the bottom four bits, taking the 0 and 1 cases here. For zero, we are
        // collecting a positive value, and for 1 a negative. Next we branch on the
        // value of nnn in 0nnn000x. For 0 we have a single octet following and
        // a value to be added (after negation if necessary) to ptroffset. For 1
        // we have a two octet value (most significant first).
        // Bit 1 is zero - further analysis needed
        int subcode = code / 16;  // pick up nnn
        int code1 = code % 16;
        if(code1 == 0 or code1 == 1) {
          //
          // codecase = 4 - diff change
          // If a change of a diff is needed, we
          // have the following cases for the diff:
          //              positive and less than 256
          //              negative and less than 256
          // We use 00100000 and 00100001 for these cases for diffa, 00110000 and
          // 00110001 for these cases for diffb, 01000000 and 01000001 for diffc,
          // and 0101000 and 01010001 for diffd, followed by a single byte which gives
          // the fcm diff value.
          // On decompression, this is the nnn = 2 (diffa), 3 (diffb), 4 (diffc) and
          // 5 (diffd) cases described under pointer above.
          //
          // This completes the use of the 0mmm0000 and 0mmm0001 values
          // except for 110 which is used in termination below, and 111 which is spare.
          // Remaining codes are
          // 0mmmxxxx where xxxx is above 0001.

          negate = code1;     // 1 if we are to negate
          switch(subcode) {
            case 0:
              octets = 1;
              codecase = 3;   // Pointer change
              break;
            case 1:
              octets = 2;
              codecase = 3;   // Pointer change
              break;
            case 2:
              codecase = 4;   // diffa change
              diffindx = 1;
              break;
            case 3:
              codecase = 4;   // diffb change
              diffindx = 2;
              break;
            case 4:
              codecase = 4;   // diffc change
              diffindx = 3;
              break;
            case 5:
              codecase = 4;   // diffd change
              diffindx = 4;
              break;
            case 6:
              //
              // codecase = 5
              // Where we have discarded some old file and refilled the buffers, we
              // signal this
              // in the output by putting out a single octet of 01100000 followed by
              // two octets giving the amount of the move.
              // On decompression, this is a signal to move the new file by the full
              // 10K, and to move the old file by the specified amount.
              //
              // codecase = 6
              // Finally, we have to cope with the residual set of up to 15 octets at the
              // end of the file. We simply output 01100001, then a single octet saying
              // how many follow, then the octets. If there are none, we still output
              // the single octet and the null count. On decompression, we copy octets
              // across.
              if(code1 == 0)
                codecase = 5;   // Move the reference file down
              else
                codecase = 6;   // Copy the residual octets from the fcm file
              break;
            case 7:
              cout << "**** ERROR - Data format error - Value 6 or 7 with code1 = 0 or 1";
              exit(-1);  // Do nothing - value 7 is spare.
          }
        }
        else {
          // Now we have cases of code1 above 0 and 1, with the subcode holding
          // a count.
          count = subcode;
          if(count == 0)
            count = 8;
          switch(code1) {
            //
            // codecase = 7
            // For a type B match, we determine whether to use diffa, b, c, or d (based
            // on this segment alone, and possibly with a diff change). We then merge in
            // up to 8 segments (nnn = 000 means 8) with the same pointer and diff value
            // (after juggling diffs if necessary). The coding is 0nnn0010 for use of diffa
            // 0nnn0011 for use of diffb, 0nnn0100 for use of diffc, and 0nnn0101 for use of
            // diffd. nnn is the number of segments merged in. We then follow with a
            // string of bits, one per byte in each of the segments, where 0 means no
            // addition, and 1 means add in the selected diff. Bits up to the next octet
            // boundary are ignored.
            // On decompression, we use nnn to determine the number of octets to be
            // produced (16 * nnn), and then take as many octets as necessary to give
            // us the additions of diffa, b etc. We start the old at cnt + ptroffset + 1
            // and we set new to old + diff. (Diff was defined as new - old)
            //
            case 2:
              codecase = 7;   // Type B
              diffindx = 1;   // use diffa
              break;
            case 3:
              codecase = 7;
              diffindx = 2;   // use diffb
              break;
            case 4:
              codecase = 7;
              diffindx = 3;   // use diffc
              break;
            case 5:
              codecase = 7;
              diffindx = 4;   // use diffd
              break;
            //
            // codecase = 8
            // For a type C match, we have six possible diff selections to use, coded as
            // follows (again, merging in up to 8 possible segments):
            //            diffa and diffb    0nnn0110
            //            diffa and diffc    0nnn0111
            //            diffa and diffd    0nnn1000
            //            diffb and diffc    0nnn1001
            //            diffb and diffd    0nnn1010
            //            diffc and diffd    0nnn1011

            // We then follow with a string of bits, one per byte in each of the segments,
            // where 0 means no addition, and 10 means the first mentioned diff is added in,
            // and 11 means the second mentionned diff is added in.
            // On decompression, this is values 6 to 11 of the bottom four octets. We
            // proceed as for case B, except that we need to use two diffs, diff1 and diff2
            // taken from diffa to diffd according to the case being considered.
            //
            case 6:
              codecase = 8;
              diffindx1 = 1;
              diffindx2 = 2;
              break;
            case 7:
              codecase = 8;
              diffindx1 = 1;
              diffindx2 = 3;
              break;
            case 8:
              codecase = 8;
              diffindx1 = 1;
              diffindx2 = 4;
              break;
            case 9:
              codecase = 8;
              diffindx1 = 2;
              diffindx2 = 3;
              break;
            case 10:
              codecase = 8;
              diffindx1 = 2;
              diffindx2 = 4;
              break;
            case 11:
              codecase = 8;
              diffindx1 = 3;
              diffindx2 = 4;
              break;
            //
            // codecases 9 and 10
            // For a type D match, we encode the segments
            // as 0nnn1100 (for up to 8 segments) where we have four differs, and
            // 0nnn1101 where diffa is omitted, 0nnn1101 where diffb is omitted, 0nnn1110
            // where diffc is omitted, and 0nnn1111 where diffd is omitted.
            // This is followed by a bitstring for each
            // octet in each of the segments, where we have 0 if no diff is to
            // be added in, then, for the three differs in use,
            // 100 if the first is to be added in, 101 for the second, and 11 for the third.
            // For four differs, we have
            // 100 if diffa is to be added in, 101 for diffb, 110 for diffc,
            // 111 for diffd.
            // On decompression, this is again similar to CASE C, except that we have
            // three or four differs depending on the value 12 (four) or 13 to 16 (three)
            // of the case. The value of the case says which differ is omitted.
            //
            case 12:
              codecase = 9;
              diffindx1 = 1;
              diffindx2 = 2;
              diffindx3 = 3;
              diffindx4 = 4;
              break;
            case 13:
              codecase = 10;
              diffindx1 = 2;
              diffindx2 = 3;
              diffindx3 = 4;
              break;
            case 14:
              codecase = 10;
              diffindx1 = 1;
              diffindx2 = 3;
              diffindx3 = 4;
              break;
            case 15:
              codecase = 10;
              diffindx1 = 1;
              diffindx2 = 2;
              diffindx3 = 4;
              break;
            // case 16 cannot occur because of lack of code space
          }
        }
      }

      if(trace) {
        dbg << "Attempting codecase " << codecase << ", Ptroffset " << ptroffset << ", Cnt " << cnt << ", code " << code << "\n";
        dbg << "           Differences: " << gdiff[1] << ", " << gdiff[2] << ", " << gdiff[3] << ", " << gdiff[4] << "\n";
        dbg.flush();
      }

      switch(codecase) {

        case 1: // Copy ISEG * count octets from the fcm to the target.
          for(ijl=1; ijl<=ISEG*count; ijl++)
            outputsub(bytesfcm[fcmpointer + ijl]);
          fcmpointer += 1 + ISEG * count;
          cnt += ISEG * count;
          break;

        case 2: // Copy ISEG * count octets from the old file, starting at
                // cnt + ptroffset + 1.
          for(ijl=1; ijl<=ISEG*count; ijl++)
            outputsub(bytesold[cnt + ptroffset + ijl]);
          fcmpointer++;
          cnt += ISEG * count;
          break;

        case 3: // Add to ptroffset, a one octet value (octets = 1) or a two
                // octet value (octets = 2), or subtract if negate = 1
          octval = bytesfcm[fcmpointer + 1];
          if(octets == 2)
            octval = octval * 256 + bytesfcm[fcmpointer + 2];
          if(octets == 1 and octval == 0)
            octval = 256;
          if(negate == 1)
            octval = -octval;
          ptroffset += octval;
          fcmpointer += octets + 1;
          break;

        case 4: // Set gdiff(diffindx) to the one octet value which follows,
          // negated if negate is one.
          octval = bytesfcm[fcmpointer + 1];
          if(negate == 1)
            octval = -octval;
          gdiff[diffindx] = octval;
          fcmpointer += 2;
          break;

        case 5: // The next two octets code up the move down of old.

          progress();

          // Now we have to move the old and adjust pointers
          move = bytesfcm[fcmpointer + 1] * 256 + bytesfcm[fcmpointer + 2];
          fcmpointer += 3;

          // move oldarray down by move and adjust cnt
          cnt -= move;
          endold -= move;
          usedold += move;
          memmove(bytesold+1, bytesold+move+1, endold);

          if(trace) {
            dbg << "Moving old file by " << move << "\n";
            dbg.flush();
          }

          // Now read in as much as possible of the old file to fill the buffer.

          memset(buff, ' ', IBUF);
          ICNT1 = endold;
          while(not fpold.eof() and ICNT1 + IBUF <= ISIZE) {
            fpold.read(buff, IBUF);
            memcpy(bytesold+ICNT1+1, buff, IBUF);
            ICNT1 += IBUF;
          }

          endold = ISIZE;
          if(lengthold - usedold < ISIZE)
            endold = (int)(lengthold - usedold);
          break;

        case 6: // Copy a number of octets given by the next octet from the
                // fcm file to the new file. This is the last up to 15 octets, and
                // we are finished. The fcm file should be exhausted.
          reallyfinished = true;
          copybytes = bytesfcm[fcmpointer + 1];
          for(ijl=2; ijl<=copybytes+1; ijl++)
            outputsub(bytesfcm[fcmpointer + ijl]);
          fcmpointer += 2 + copybytes;
          cnt += copybytes;
          break;

        case 7: // Produce count * ISEG octets of new file by taking the old
                // and adding gdiff(diffindx) if the bit in the following octets is one.
          bitstringptr = 0;
          fcmpointer++;
          numocts = count * ISEG;
          while(numocts != 0) {
            if(getbit() == 0)
              outputsub(bytesold[cnt + ptroffset + 1]);
            else
              outputsub(bytesold[cnt + ptroffset + 1] + gdiff[diffindx]);
            numocts--;
            cnt++;
          }
          break;

        case 8: // Produce count * ISEG octets of new file by taking the old
                // and adding gdiff(diffindx1) for 10, and gdiff(diffindx2) for 11, and
                // nothing for 0
          bitstringptr = 0;
          fcmpointer++;
          numocts = count * ISEG;
          while(numocts != 0) {
            if(getbit() == 0)
              outputsub(bytesold[cnt + ptroffset + 1]);
            else {
              if(getbit() == 0)
                outputsub(bytesold[cnt + ptroffset + 1] + gdiff[diffindx1]);
              else
                outputsub(bytesold[cnt + ptroffset + 1] + gdiff[diffindx2]);
            }
            numocts--;
            cnt++;
          }
          break;

        case 9: // Produce count * ISEG octets of new file by taking the old
                // and adding nothing for 0, gdiff(1) for 100, gdiff(2) for 101, gdiff(3)
                // for 110, and gdiff(4) for 111.
          bitstringptr = 0;
          fcmpointer++;
          numocts = count * ISEG;
          while(numocts != 0) {
            if(getbit() == 0)
              outputsub(bytesold[cnt + ptroffset + 1]);
            else {
              if(getbit() == 0) {
                if(getbit() == 0)
                  outputsub(bytesold[cnt + ptroffset + 1] + gdiff[diffindx1]);
                else
                  outputsub(bytesold[cnt + ptroffset + 1] + gdiff[diffindx2]);
              }
              else {
                if(getbit() == 0)
                  outputsub(bytesold[cnt + ptroffset + 1] + gdiff[diffindx3]);
                else
                  outputsub(bytesold[cnt + ptroffset + 1] + gdiff[diffindx4]);
              }
            }
            numocts--;
            cnt++;
          }
          break;

        case 10: // Produce count * ISEG octets of new file by taking the old
                 // and adding nothing for 0, gdiff(diffindx1) for 100, and gdiff(diffindx2)
                 // for 101, gdiff(diffindx3) for 11.
          bitstringptr = 0;
          fcmpointer = fcmpointer + 1;
          numocts = count * ISEG;
          while(numocts != 0) {
            if(getbit() == 0)
              outputsub(bytesold[cnt + ptroffset + 1]);
            else {
              if(getbit() == 0) {
                if(getbit() == 0)
                  outputsub(bytesold[cnt + ptroffset + 1] + gdiff[diffindx1]);
                else
                  outputsub(bytesold[cnt + ptroffset + 1] + gdiff[diffindx2]);
              }
              else {
                outputsub(bytesold[cnt + ptroffset + 1] + gdiff[diffindx3]);
              }
            }
            numocts--;
            cnt++;
          }
          break;

        default:
          cout << "**** ERROR - Data format error - Bad codecase" << endl;
          exit(-1);

      }
    }

    // Now loop to do another hunk if necessary
    finished = false;
    if(lengthfcm <= endfcm + donefcm) {
      finished = true;
    }
    else {

      progress();

      // Prepare for next hunk

      fcmpointer = fcmpointer - ISIZE3;
      donefcm = donefcm + ISIZE3;

      // move fcm array down by ISIZE3
      realendfcm -= ISIZE3;
      memmove(bytesfcm+ijl, bytesfcm+ijl+ISIZE3, realendfcm);

      // Now read in as much as possible of the fcm file to fill the buffer.

      memset(buff, ' ', IBUF);
      ICNT2 = realendfcm;
      while(not fpfcm.eof() and ICNT2 + IBUF <= ISIZE) {
        fpfcm.read(buff, IBUF);
        memcpy(bytesold+ICNT2+1, buff, IBUF);
        ICNT2 += IBUF;
      }

      realendfcm = ISIZE;
      endfcm = ISIZE2;
      if(lengthfcm - donefcm <= ISIZE) {
        endfcm = (int)(lengthfcm - donefcm);
        realendfcm = endfcm;
      }

      // This ends the preparation for the next hunk.
    }

  } while(not finished);

  if(not reallyfinished) {
    cout <<  "**** ERROR - Data format error - Ran out of difference file before finished" << endl;
    remove(wrkfile);
    exit(-1);
  }

  fp.write(obuff, obufflen);

  progress();

  fpfcm.close();
  fpold.close();
  fp.close();

  if(sumcheckval1 != sumcheck1 or sumcheckval2 != sumcheck2) {
    cout << "**** ERROR - Sumcheck of reconstituted file differs from original" << endl;
    remove(wrkfile);
    exit(-1);
  }
  else {
    remove(newfile);
    rename(wrkfile, newfile);
    if(gdffmt) {
      struct utimbuf t;
      t.actime = t.modtime = newstamp;
      utime(newfile, &t);
    }

    if(not quiet)
      cout << endl << endl << "Done." << endl;
  }

  return 0;
}


//  ------------------------------------------------------------------

